﻿-- function schedluer frame
local BMTSchedulerFrame = CreateFrame("Frame");
-- time since last call of the OnUpdate
local BMTElapsed = 0;
BMTScheduledTasks = {};
-- OnUpdate calls throttle
local BMT_ONUPDATE_SEC = 0.5;
-- set to true for additional dev messages
local BMTDebug = false;
-- title text on top of the browser window
local BMT_TITLE = "Black Market Tracker";
-- new addon version text shown at the bottom of the browser
local BMT_NEW_VERSION = "New addon version available";
-- coming soon text
local BMT_COMING_SOON = "Coming Soon!";
-- alpa for fading bottom text, goes above 1.0 so that the text stays lit for a while
local BMT_TEXT_ALPHA = 6.0;
-- time of fade in/out
local BMT_TEXT_TIME = 2.5;
-- sync string separator
local BMT_SEPARATOR = "%";
-- sync sender (through new search) key
local BMT_SYNC_SENDER_KEY = "UPDATER";
-- request update times key
local BMT_SYNC_TIMEREQ_KEY = "TIMEREQ";
-- response update times key
local BMT_SYNC_TIMERESP_KEY = "TIMERESP";
-- sync item key
local BMT_SYNC_ITEM_KEY = "ITEM";
-- sync item request key
local BMT_SYNC_ITEMREQ_KEY = "ITEMREQ";
-- sync item end key
local BMT_SYNC_ITEMEND_KEY = "ITEMEND";
-- sync version reqiest key
local BMT_SYNC_VERREQ_KEY = "VERREQ";
-- sync version response key
local BMT_SYNC_VERRESP_KEY = "VERRESP";
-- time for syncing the update times upon logon, in seconds
local BMT_SYNC_TIME_DELAY = 7;
-- time for syncing the items, in seconds
local BMT_SYNC_ITEMS_DELAY = 3;
-- predicted time left constants (time in minutes) - completed, short, medium, long, very long
local BMT_TIME_LEFT = {0, 30, 120, 720, 9999};
-- this revision number
local BMTRevision = GetAddOnMetadata("BlackMarketTracker", "Version");
-- highest revision gotten from others upon logging in
BMTHighestRevision = BMTRevision;
-- fading state (bottom text)
BMTFadeState = 0;
-- last updated time string, used to throttle the sync results (when multiple events get fired after opening BMAH window once)
BMTLastUpdateTime = "-";
-- true if currently syncing items on logon, false otherwise. When false, BMT_SYNC_TIMERESP_KEY is ignored
BMTSyncingTime = false;
-- best sync time
BMTBestUpdateTime = 0;
-- name of the person with best sync time
BMTBestUpdateTimeSender = nil;
-- item database from syncingBMTPredictedLow
BMTSyncItemDB = nil;

function BMT_Debug(msg)
	if(BMTDebug) then
		DEFAULT_CHAT_FRAME:AddMessage(msg);
	end
end

function BMT_SlashDebugHandler()
end

BMTSchedulerFrame:SetScript("OnUpdate", function(self, elapsedVal)
	-- throttle BMT_OnUpdate calls
	BMTElapsed = BMTElapsed + elapsedVal;
	if(BMTElapsed >= BMT_ONUPDATE_SEC) then
		BMTElapsed = 0;
		BMT_OnUpdate(self, elapsedVal);
	end
end)

function BMT_SlashDeleteHandler()
	DEFAULT_CHAT_FRAME:AddMessage("Black Market Tracker: Deleting all data.");
	BlackMarketTrackerCurrentDB = nil;
	BlackMarketTrackerHistoryDB = nil;
	BlackMarketTrackerConfig = nil;
	BlackMarketTrackerMisc = nil;
end

function BMT_SlashHandler(msg, editbox)
	local msgLen = #msg;
	-- TODO: Config
	if(msgLen == 0 or msg == "show" or msg == "ui" or msg == "open") then
		BMTFrame_Toggle();
	elseif(msg == "debug") then
		BMT_SlashDebugHandler();
	elseif(msg == "delete") then
		BMT_SlashDeleteHandler();
	end
end

function BMTFrame_OnLoad(self)
	SLASH_BLACKMARKETTRACKER1 = "/bmt";
	SLASH_BLACKMARKETTRACKER2 = "/blackmarkettracker";
	SlashCmdList["BLACKMARKETTRACKER"] = BMT_SlashHandler;
	self:RegisterForDrag("LeftButton");
	tinsert(UISpecialFrames, self:GetName());
	self:RegisterEvent("PLAYER_LOGIN");
	self:RegisterEvent("CHAT_MSG_ADDON");
	self:RegisterEvent("BLACK_MARKET_ITEM_UPDATE");
	BlackMarketTrackerScrollFrame.scrollBar.doNotHide = true;
	BlackMarketTrackerFrameTitleText:SetText(BMT_TITLE);
	RegisterAddonMessagePrefix("BMT");
end

function BMTFrame_OnEvent(self, event, ...)
	if(event == "BLACK_MARKET_ITEM_UPDATE") then
		BMTTaskManager_Unschedule(BMT_NewSearchHandler);
		BMTTaskManager_Schedule(0.5, BMT_NewSearchHandler);
	elseif(event == "CHAT_MSG_ADDON") then
		local prefix = select(1, ...);
		local msg = select(2, ...);
		local channel = select(3, ...);
		local sender = select(4, ...);
		BMT_AddonMsgReceiveHandler(prefix, msg, channel, sender);
	elseif(event == "PLAYER_LOGIN") then
		BMT_OnLoginHandler();
	end
end

function BMTFrame_Toggle()
	if(BlackMarketTrackerFrame:IsShown()) then
		BlackMarketTrackerFrame:Hide();
		--BlackMarketTrackerBottom:SetAlpha(0);
		-- TODO: sound enable toggle
	else
		BlackMarketTrackerFrame:Show();
		-- TODO: sound enable toggle
	end
end

function BMTFrame_OnShow()
	if(BMTHighestRevision <= BMTRevision) then
		BMTFadeState = 2;
	else
		BMTFadeState = 0;
	end
	PlaySound("AuctionWindowOpen");
	HybridScrollFrame_CreateButtons(BlackMarketTrackerScrollFrame, "BlackMarketTrackerItemTemplate", 5, -5);
	BlackMarketTrackerScrollFrame_Update();
	BMTBottomText_FadingHandler();
end

-- delay_time, function, function_parameters
function BMTTaskManager_Schedule(delay, func, ...)
	-- add params
	local task = {...};
	-- add function and time when it should be called
	task.func = func;
	task.time = GetTime() + delay;
	table.insert(BMTScheduledTasks, task);
end

function BMTTaskManager_Unschedule(func, ...)
	local entry, found;
	for i=#BMTScheduledTasks, 1, -1 do
		entry = BMTScheduledTasks[i];
		if(entry.func == func) then
			found = true;
			for j=1, select("#", ...) do
				if(select(j, ...) ~= entry[i]) then
					found = false;
					break;
				end
			end
			if(found) then
				table.remove(BMTScheduledTasks, i);
			end
		end
	end
end

function BMT_OnUpdate(self, elapsedVal)
	-- for some reason this needs to be done
	if(BMTElapsed == 0) then return end
	
	--BMT_Debug(GetTime().." BMTElapsed="..BMTElapsed);
	local entry;
	for i=#BMTScheduledTasks, 1, -1 do
		entry = BMTScheduledTasks[i];
		if(entry.time <= GetTime()) then
			table.remove(BMTScheduledTasks, i);
			entry.func(unpack(entry));
		end
	end
end

function BMTItem_OnClick(self, button, down)
	-- Function for later use
end

function BMTBottomText_FadingHandler()
	if(BMTFadeState == 0) then
		BMTFadeState = 1;
		BlackMarketTrackerFrameBottomText:SetText(BMT_NEW_VERSION .. " (r" .. BMTHighestRevision .. ")");
		BMTTrackerBottomText_FadeIn();
	elseif(BMTFadeState == 1) then
		BMTFadeState = 2;
		BMTBottomText_FadeOut();
	elseif(BMTFadeState == 2) then
		BMTFadeState = 3;
		local bottomText = "Last updated by ";
		if(BlackMarketTrackerMisc) then
			if(not BlackMarketTrackerMisc.UpdaterName or not BlackMarketTrackerMisc.UpdaterClass or not BlackMarketTrackerMisc.UpdaterGuild or not BlackMarketTrackerMisc.UpdateTime or not BlackMarketTrackerMisc.UpdateDate) then
				bottomText = bottomText .. "|cffffd100Unknown|r";
			else
				local classColor = RAID_CLASS_COLORS[BlackMarketTrackerMisc.UpdaterClass];
				bottomText = bottomText .. "|cff" .. string.format("%2.2x", classColor.r*255) .. string.format("%2.2x", classColor.g*255) .. string.format("%2.2x", classColor.b*255);
				bottomText = bottomText .. BlackMarketTrackerMisc.UpdaterName .. "|r (";
				bottomText = bottomText .. BlackMarketTrackerMisc.UpdaterGuild .. "), ";
				local updateTimeStr =  string.format("%2.2d:%2.2d", (BlackMarketTrackerMisc.UpdateTime/60), (BlackMarketTrackerMisc.UpdateTime%60));
				bottomText = bottomText .. BlackMarketTrackerMisc.UpdateDate .. " " .. updateTimeStr;
			end
		else
			bottomText = bottomText .. "|cffffd100Unknown|r";
		end
		BlackMarketTrackerFrameBottomText:SetText(bottomText);
		BMTTrackerBottomText_FadeIn();
	end
end

function BMTTrackerBottomText_FadeIn()
	local fadeInfo = {};
    fadeInfo.mode = "IN";
    fadeInfo.timeToFade = BMT_TEXT_TIME;
	fadeInfo.endAlpha = BMT_TEXT_ALPHA;
    fadeInfo.finishedFunc = BMTBottomText_FadingHandler;
    UIFrameFade(BlackMarketTrackerFrameBottom, fadeInfo);
end

function BMTBottomText_FadeOut()
	local fadeInfo = {};
    fadeInfo.mode = "OUT";
    fadeInfo.timeToFade = BMT_TEXT_TIME;
	fadeInfo.startAlpha = BMT_TEXT_ALPHA;
    fadeInfo.finishedFunc = BMTBottomText_FadingHandler;
    UIFrameFade(BlackMarketTrackerFrameBottom, fadeInfo);
end

function BlackMarketTrackerScrollFrame_Update()
	local numItems = 0;
	if(BlackMarketTrackerCurrentDB ~= nil) then 
		numItems = #BlackMarketTrackerCurrentDB;
	end
	
	local scrollFrame = BlackMarketTrackerScrollFrame;
	local offset = HybridScrollFrame_GetOffset(scrollFrame);
	local buttons = scrollFrame.buttons;
	local numButtons = #buttons;
	
	for i=1, numButtons do
		local button = buttons[i];
		-- TODO: (should be history db as well) get index of an item in black market
		local index = offset + i;
		if ( index <= numItems ) then
			local name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID = BMT_GetFromCurrentDB(index);
			if(name) then
				button.Name:SetText(name);
                button.Item.IconTexture:SetTexture(texture);
				--[[
                if ( not usable ) then
                    button.Item.IconTexture:SetVertexColor(1.0, 0.1, 0.1);
                else
                    button.Item.IconTexture:SetVertexColor(1.0, 1.0, 1.0);
                end
				]]
                button.Item.Count:SetText(quantity);
                button.Item.Count:SetShown(quantity > 1);
                button.Type:SetText(itemType);
                button.Seller:SetText(sellerName);
                button.Level:SetText(level);
                local bidAmount = currBid;
                local minNextBid = currBid + minIncrement;
                if ( currBid == 0 ) then
                    bidAmount = minBid;
                    minNextBid = minBid;
                end
                MoneyFrame_Update(button.CurrentBid, bidAmount);
                button.minNextBid = minNextBid;
                button.auctionCompleate = (timeLeft == 0);
                button.TimeLeft.Text:SetText(_G["AUCTION_TIME_LEFT"..timeLeft]);
                button.TimeLeft.tooltip = _G["AUCTION_TIME_LEFT"..timeLeft.."_DETAIL"];
                button.itemLink = link;
				button.PredictedTimeLeft:SetText(BMT_COMING_SOON);
				button.Warning:SetText(BMT_COMING_SOON);
                button:Show();
            else
                button:Hide();
			end
		else
			button:Hide();
		end
	end
	
	local totalHeight = numItems * scrollFrame.buttonHeight;
    local displayedHeight = numButtons * scrollFrame.buttonHeight;
    HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);
end

function BMT_SetNewUpdater(updaterName, updaterClass, updaterGuild, updateTime, updateDate)
	if (not BlackMarketTrackerMisc) then
		BlackMarketTrackerMisc = {};
	else
		-- throttle
		local currHour, currMinute = GetGameTime();
		local currTime = 60*currHour+currMinute;
		if(currTime == BlackMarketTrackerMisc.UpdateTime) then return end
	end
	
	BlackMarketTrackerMisc.UpdaterName = updaterName;
	BlackMarketTrackerMisc.UpdaterClass = updaterClass;
	BlackMarketTrackerMisc.UpdaterGuild = updaterGuild;
	BlackMarketTrackerMisc.UpdateTime = updateTime;
	BlackMarketTrackerMisc.UpdateDate = updateDate;
	
	-- if tracker is opened already, update the bottom text
	if(BlackMarketTrackerFrame:IsShown()) then
		BMTFadeState = 2;
		BMTBottomText_FadingHandler();
	end
end

function BMT_AddNewToCurrentDB(name, texture, quantity, itemType,  level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID)
	-- check if table doesn't exist should already be done
	local currDBSize = #BlackMarketTrackerCurrentDB;
	local currHour, currMinute = GetGameTime();
	local currTime = 60*currHour+currMinute;
	BlackMarketTrackerCurrentDB[currDBSize+1] = {};
	BlackMarketTrackerCurrentDB[currDBSize+1].updateTime = currTime;
	BlackMarketTrackerCurrentDB[currDBSize+1].name = name;
	BlackMarketTrackerCurrentDB[currDBSize+1].texture = texture;
	BlackMarketTrackerCurrentDB[currDBSize+1].quantity = quantity;
	BlackMarketTrackerCurrentDB[currDBSize+1].itemType = itemType;
	BlackMarketTrackerCurrentDB[currDBSize+1].level = level;
	BlackMarketTrackerCurrentDB[currDBSize+1].levelType = levelType;
	BlackMarketTrackerCurrentDB[currDBSize+1].sellerName = sellerName;
	BlackMarketTrackerCurrentDB[currDBSize+1].minBid = minBid;
	BlackMarketTrackerCurrentDB[currDBSize+1].minIncrement = minIncrement;
	BlackMarketTrackerCurrentDB[currDBSize+1].currBid = currBid;
	-- TODO: Use for highest bidder detection
	BlackMarketTrackerCurrentDB[currDBSize+1].youHaveHighBid = youHaveHighBid;
	-- TODO: Check field
	BlackMarketTrackerCurrentDB[currDBSize+1].numBids = numBids;
	BlackMarketTrackerCurrentDB[currDBSize+1].timeLeft = timeLeft;
	-- completed
	if(timeLeft == 0) then
		BlackMarketTrackerCurrentDB[currDBSize+1].predictedLow = 0;
		BlackMarketTrackerCurrentDB[currDBSize+1].predictedHigh = 0;
	else
		BlackMarketTrackerCurrentDB[currDBSize+1].predictedLow = BMT_TIME_LEFT[timeLeft];
		BlackMarketTrackerCurrentDB[currDBSize+1].predictedHigh = BMT_TIME_LEFT[timeLeft+1];
	end
	BlackMarketTrackerCurrentDB[currDBSize+1].link = link;
	BlackMarketTrackerCurrentDB[currDBSize+1].marketID = marketID;
end

function BMT_UpdateExistingCurrentDBEntry(index, name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID)
	local currHour, currMinute = GetGameTime();
	local currTime = 60*currHour+currMinute;
	-- throttle the updates
	if(currTime == BlackMarketTrackerCurrentDB[index].updateTime) then return end
	BlackMarketTrackerCurrentDB[index].updateTime = currTime;
	BlackMarketTrackerCurrentDB[index].name = name;
	BlackMarketTrackerCurrentDB[index].texture = texture;
	BlackMarketTrackerCurrentDB[index].quantity = quantity;
	BlackMarketTrackerCurrentDB[index].itemType = itemType;
	BlackMarketTrackerCurrentDB[index].level = level;
	BlackMarketTrackerCurrentDB[index].levelType = levelType;
	BlackMarketTrackerCurrentDB[index].sellerName = sellerName;
	BlackMarketTrackerCurrentDB[index].minBid = minBid;
	BlackMarketTrackerCurrentDB[index].minIncrement = minIncrement;
	BlackMarketTrackerCurrentDB[index].currBid = currBid;
	-- TODO: Use for highest bidder detection
	BlackMarketTrackerCurrentDB[index].youHaveHighBid = youHaveHighBid;
	-- TODO: Check field
	BlackMarketTrackerCurrentDB[index].numBids = numBids;
	BlackMarketTrackerCurrentDB[index].timeLeft = timeLeft;
	-- TODO: Recalculate predicted time
	if(timeLeft == 0) then
		BlackMarketTrackerCurrentDB[index].predictedLow = 0;
		BlackMarketTrackerCurrentDB[index].predictedHigh = 0;
	else
		BlackMarketTrackerCurrentDB[index].predictedLow = BMT_TIME_LEFT[timeLeft];
		BlackMarketTrackerCurrentDB[index].predictedHigh = BMT_TIME_LEFT[timeLeft+1];
	end
	BlackMarketTrackerCurrentDB[index].link = link;
	BlackMarketTrackerCurrentDB[index].marketID = marketID;
end

function BMT_GetFromCurrentDB(index)
	return BlackMarketTrackerCurrentDB[index].name, BlackMarketTrackerCurrentDB[index].texture, BlackMarketTrackerCurrentDB[index].quantity, BlackMarketTrackerCurrentDB[index].itemType, BlackMarketTrackerCurrentDB[index].level, BlackMarketTrackerCurrentDB[index].levelType, BlackMarketTrackerCurrentDB[index].sellerName, BlackMarketTrackerCurrentDB[index].minBid, BlackMarketTrackerCurrentDB[index].minIncrement, BlackMarketTrackerCurrentDB[index].currBid, BlackMarketTrackerCurrentDB[index].youHaveHighBid, BlackMarketTrackerCurrentDB[index].numBids, BlackMarketTrackerCurrentDB[index].timeLeft, BlackMarketTrackerCurrentDB[index].link, BlackMarketTrackerCurrentDB[index].marketID, BlackMarketTrackerCurrentDB[index].predictedLow, BlackMarketTrackerCurrentDB[index].predictedHigh;
end

function BMT_UpdateCurrentDB(name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID, predictedLow, predictedHigh)
	predictedLow = predictedLow or -1;
	predictedHigh = predictedHigh or -1;
	if(not BlackMarketTrackerCurrentDB) then
		BlackMarketTrackerCurrentDB = {};
	end
	local currDBSize = #BlackMarketTrackerCurrentDB;
	-- iterate through all fields, check for duplicates
	local duplicateIndex = 0;
	for i=1, currDBSize do
		if(BlackMarketTrackerCurrentDB[i].name == name) then
			duplicateIndex = i;
		end
	end
	-- new entry
	if(duplicateIndex == 0) then
		BMT_AddNewToCurrentDB(name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID);
	-- existing entry
	else
		BMT_UpdateExistingCurrentDBEntry(duplicateIndex, name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID);
	end
end

function BMT_MoveEntryToHistoryDB(index)
	if(not BlackMarketTrackerHistoryDB) then
		BlackMarketTrackerHistoryDB = {};
	end
	BMT_Debug("Moving "..BlackMarketTrackerCurrentDB[index].name.." to history db");
	local histDBSize = #BlackMarketTrackerHistoryDB;
	BlackMarketTrackerHistoryDB[histDBSize+1] = {};
	BlackMarketTrackerHistoryDB[histDBSize+1].name = BlackMarketTrackerCurrentDB[index].name;
	BlackMarketTrackerHistoryDB[histDBSize+1].texture = BlackMarketTrackerCurrentDB[index].texture;
	BlackMarketTrackerHistoryDB[histDBSize+1].quantity = BlackMarketTrackerCurrentDB[index].quantity;
	BlackMarketTrackerHistoryDB[histDBSize+1].itemType = BlackMarketTrackerCurrentDB[index].itemType;
	BlackMarketTrackerHistoryDB[histDBSize+1].level = BlackMarketTrackerCurrentDB[index].level;
	BlackMarketTrackerHistoryDB[histDBSize+1].levelType = BlackMarketTrackerCurrentDB[index].levelType;
	BlackMarketTrackerHistoryDB[histDBSize+1].sellerName = BlackMarketTrackerCurrentDB[index].sellerName;
	BlackMarketTrackerHistoryDB[histDBSize+1].minBid = BlackMarketTrackerCurrentDB[index].minBid;
	BlackMarketTrackerHistoryDB[histDBSize+1].minIncrement = BlackMarketTrackerCurrentDB[index].minIncrement;
	BlackMarketTrackerHistoryDB[histDBSize+1].currBid = BlackMarketTrackerCurrentDB[index].currBid;
	BlackMarketTrackerHistoryDB[histDBSize+1].youHaveHighBid = BlackMarketTrackerCurrentDB[index].youHaveHighBid;
	BlackMarketTrackerHistoryDB[histDBSize+1].numBids = BlackMarketTrackerCurrentDB[index].numBids;
	BlackMarketTrackerHistoryDB[histDBSize+1].timeLeft = 0;
	BlackMarketTrackerHistoryDB[histDBSize+1].link = BlackMarketTrackerCurrentDB[index].link;
	BlackMarketTrackerHistoryDB[histDBSize+1].marketID = BlackMarketTrackerCurrentDB[index].marketID;
	table.remove(BlackMarketTrackerCurrentDB, index);
end

function BMT_SendCurrentDB()
	if(not BlackMarketTrackerCurrentDB or not BlackMarketTrackerMisc) then return end
	if(not BlackMarketTrackerMisc.UpdaterName) then return end
	
	local currDBSize = #BlackMarketTrackerCurrentDB;
	for i=1, currDBSize do
		-- TODO: Recalculate predicted time before sending
		local name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID, predictedLow, predictedHigh = BMT_GetFromCurrentDB(i);
		SendAddonMessage("BMT", BMT_SYNC_ITEM_KEY..BMT_SEPARATOR..name..BMT_SEPARATOR..texture..BMT_SEPARATOR..quantity..BMT_SEPARATOR..itemType..BMT_SEPARATOR..level..BMT_SEPARATOR..levelType..BMT_SEPARATOR..sellerName..BMT_SEPARATOR..minBid..BMT_SEPARATOR..minIncrement..BMT_SEPARATOR..currBid..BMT_SEPARATOR..tostring(youHaveHighBid)..BMT_SEPARATOR..numBids..BMT_SEPARATOR..timeLeft..BMT_SEPARATOR..link..BMT_SEPARATOR..marketID..BMT_SEPARATOR..predictedLow..BMT_SEPARATOR..predictedHigh, "GUILD");
	end
	SendAddonMessage("BMT", BMT_SYNC_ITEMEND_KEY, "GUILD");
	SendAddonMessage("BMT", BMT_SYNC_SENDER_KEY..BMT_SEPARATOR..playerName..BMT_SEPARATOR..playerClass..BMT_SEPARATOR..playerGuild..BMT_SEPARATOR..updateTime..BMT_SEPARATOR..updateDate, "GUILD");
end

function BMT_NewSearchHandler()
	-- check if blizzard's black market ui is still opened
	if(not BlackMarketFrame:IsShown()) then return end
	local playerName = UnitName("player");
	local _, playerClass = UnitClass("player");
	local playerGuild = GetGuildInfo("player");
	local updateHour, updateMinute = GetGameTime();
	local updateTime = 60*updateHour+updateMinute;
	local updateDate = date("%d/%m/%y");
	BMT_SetNewUpdater(playerName, playerClass, playerGuild, updateTime, updateDate);
	-- TODO: Friends
	local numItems = C_BlackMarket.GetNumItems();
	if(numItems == nil) then return end
	for i=1, numItems do
		local name, texture, quantity, itemType, _, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID = C_BlackMarket.GetItemInfoByIndex(i)
		BMT_UpdateCurrentDB(name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID);
	end
	-- move to history DB
	if(BlackMarketTrackerCurrentDB) then
		local currDBSize = #BlackMarketTrackerCurrentDB;
		-- iterate through all items in db
		local found;
		local i=1;
		while(i<=currDBSize) do
			found = 0;
			-- iterate through bmah items
			for j=1, numItems do
				local name, texture, quantity, itemType, _, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID = C_BlackMarket.GetItemInfoByIndex(j);
				if(BlackMarketTrackerCurrentDB[i].name == name) then
					found = 1;
				end
			end
			if(found == 0) then
				BMT_MoveEntryToHistoryDB(i);
				i = i-1;
				currDBSize = currDBSize-1;
			end
			i = i+1;
		end
	end
	-- sync to others
	BMT_SendCurrentDB();
end

function BMT_SyncSenderUpdateHandler(msg)
	local _, updaterName, updaterClass, updaterGuild, updateTime, updateDate = strsplit(BMT_SEPARATOR, msg);
	BMT_SetNewUpdater(updaterName, updaterClass, updaterGuild, updateTime, updateDate);
end

-- TODO: highest bidder
function BMT_SyncItemUpdateHandler(msg)
	local _, name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID, predictedLow, predictedHigh = strsplit(BMT_SEPARATOR, msg);
	-- refresh the delay
	BMTTaskManager_Unschedule(BMT_SYNC_ITEMS_DELAY, BMT_FinishSyncingItems);
	BMTTaskManager_Schedule(BMT_SYNC_ITEMS_DELAY, BMT_FinishSyncingItems);
	if(not BMTSyncItemDB) then
		BMTSyncItemDB = {};
	end
	local syncDBSize = #BMTSyncItemDB;
	BMTSyncItemDB[syncDBSize+1] = {};
	BMTSyncItemDB[syncDBSize+1].name = name;
	BMTSyncItemDB[syncDBSize+1].texture = texture;
	BMTSyncItemDB[syncDBSize+1].quantity = tonumber(quantity);
	BMTSyncItemDB[syncDBSize+1].itemType = itemType;
	BMTSyncItemDB[syncDBSize+1].level = tonumber(level);
	BMTSyncItemDB[syncDBSize+1].levelType = levelType;
	BMTSyncItemDB[syncDBSize+1].sellerName = sellerName;
	BMTSyncItemDB[syncDBSize+1].minBid = tonumber(minBid);
	BMTSyncItemDB[syncDBSize+1].minIncrement = tonumber(minIncrement);
	BMTSyncItemDB[syncDBSize+1].currBid = tonumber(currBid);
	-- TODO: highest bidder
	BMTSyncItemDB[syncDBSize+1].youHaveHighBid = false;
	BMTSyncItemDB[syncDBSize+1].numBids = tonumber(numBids);
	BMTSyncItemDB[syncDBSize+1].timeLeft = tonumber(timeLeft);
	BMTSyncItemDB[syncDBSize+1].link = link;
	BMTSyncItemDB[syncDBSize+1].marketID = tonumber(marketID);
	BMTSyncItemDB[syncDBSize+1].predictedLow = tonumber(predictedLow);
	BMTSyncItemDB[syncDBSize+1].predictedHigh = tonumber(predictedHigh);
end

function BMT_FinishSyncingItems()
	if(not BMTSyncItemDB) then return end
	if(not BlackMarketTrackerCurrentDB) then
		BlackMarketTrackerCurrentDB = {};
	end
	local currDBSize = #BlackMarketTrackerCurrentDB;
	local syncDBSize = #BMTSyncItemDB;
	for i=1, syncDBSize do
		BMT_UpdateCurrentDB(BMTSyncItemDB[i].name, BMTSyncItemDB[i].texture, BMTSyncItemDB[i].quantity, BMTSyncItemDB[i].itemType, BMTSyncItemDB[i].level, BMTSyncItemDB[i].levelType, BMTSyncItemDB[i].sellerName, BMTSyncItemDB[i].minBid, BMTSyncItemDB[i].minIncrement, BMTSyncItemDB[i].currBid, BMTSyncItemDB[i].youHaveHighBid, BMTSyncItemDB[i].numBids, BMTSyncItemDB[i].timeLeft, BMTSyncItemDB[i].link, BMTSyncItemDB[i].marketID, BMTSyncItemDB[i].predictedLow, BMTSyncItemDB[i].predictedHigh);
	end
	local i=1;
	while(i<=currDBSize) do
		found = 0;
		-- iterate through synced items
		for j=1, syncDBSize do
			if(BlackMarketTrackerCurrentDB[i].name == BMTSyncItemDB[j].name) then
				found = 1;
			end
		end
		if(found == 0) then
			BMT_MoveEntryToHistoryDB(i);
			i = i-1;
			currDBSize = currDBSize-1;
		end
		i = i+1;
	end
	BMTSyncItemDB = nil;
end

function BMT_SyncVersionHandler(msg)
	local key, verNumberStr = strsplit(BMT_SEPARATOR, msg);
	local verNum = tonumber(verNumberStr);
	-- this is higher version
	if(BMTRevision > verNum and key == BMT_SYNC_VERREQ_KEY) then
		SendAddonMessage("BMT", BMT_SYNC_VERRESP_KEY..BMT_SEPARATOR..BMTRevision, "GUILD");
	elseif(BMTRevision < verNum) then
		BMTHighestRevision = verNum;
		-- if ui is open, update the bottom text
		if(BlackMarketTrackerFrame:IsShown()) then
			BMTFadeState = 0;
			BMTBottomText_FadingHandler();
		end
	end
end

-- response to update time request
function BMT_SyncSendUpdateTime(msg)
	-- don't respond if this has no DB
	if(not BlackMarketTrackerMisc or not BlackMarketTrackerCurrentDB) then return end
	if(not BlackMarketTrackerMisc.UpdaterName or not BlackMarketTrackerMisc.UpdateTime) then return end
	local _, updateTimeStr = strsplit(BMT_SEPARATOR, msg);
	local updateTime = tonumber(updateTimeStr);
	if(updateTime < BlackMarketTrackerMisc.UpdateTime) then
		-- this has newer database, let others know
		SendAddonMessage("BMT", BMT_SYNC_TIMERESP_KEY..BMT_SEPARATOR..BlackMarketTrackerMisc.UpdateTime, "GUILD");
	end
end

-- received update time when logging in
function BMT_SyncUpdateTimeReceiveHandler(msg, sender)
	local _, updateTimeStr = strsplit(BMT_SEPARATOR, msg);
	local updateTime = tonumber(updateTimeStr);
	if(updateTime > BMTBestUpdateTime) then
		BMTBestUpdateTime = updateTime;
		BMTBestUpdateTimeSender = sender;
	end
end

function BMT_AddonMsgReceiveHandler(prefix, msg, channel, sender)
	if(prefix ~= "BMT") then return end
	BMT_Debug("("..sender..") msg="..msg);
	if(sender == UnitName("player")) then return end
	local key = strsplit(BMT_SEPARATOR, msg);
	if(key == BMT_SYNC_SENDER_KEY) then
		BMT_SyncSenderUpdateHandler(msg);
	elseif(key == BMT_SYNC_ITEM_KEY) then
		BMT_SyncItemUpdateHandler(msg);
	elseif(key == BMT_SYNC_VERREQ_KEY or key == BMT_SYNC_VERRESP_KEY) then
		BMT_SyncVersionHandler(msg);
	elseif(key == BMT_SYNC_TIMEREQ_KEY) then
		BMT_SyncSendUpdateTime(msg);
	elseif(key == BMT_SYNC_TIMERESP_KEY and BMTSyncingTime == true) then
		BMT_SyncUpdateTimeReceiveHandler(msg, sender);
	elseif(key == BMT_SYNC_ITEMREQ_KEY) then
		local _, name = strsplit(BMT_SEPARATOR, msg);
		if(name == UnitName("player")) then
			BMT_SendCurrentDB();
		end
	elseif(key == BMT_SYNC_ITEMEND_KEY) then
		BMTTaskManager_Unschedule(BMT_SYNC_ITEMS_DELAY, BMT_FinishSyncingItems);
		BMT_FinishSyncingItems();
	end
end

function BMT_FinishSyncingTime()
	-- ignore other responses
	BMTSyncingTime = false;
	-- check if someone had better DB
	if(BMTBestUpdateTimeSender == UnitName("player") or BMTBestUpdateTimeSender == nil) then return end
	-- request best database
	SendAddonMessage("BMT", BMT_SYNC_ITEMREQ_KEY..BMT_SEPARATOR..BMTBestUpdateTimeSender, "GUILD");
end

function BMT_OnLoginHandler()
	-- sync version
	SendAddonMessage("BMT", BMT_SYNC_VERREQ_KEY..BMT_SEPARATOR..BMTRevision, "GUILD");
	BMTSyncingTime = true;
	-- get best times
	if(BlackMarketTrackerMisc) then
		if(BlackMarketTrackerMisc.UpdateTime) then
			BMTBestUpdateTime = BlackMarketTrackerMisc.UpdateTime;
			BMTBestUpdateTimeSender = UnitName("player");
		end
	end
	SendAddonMessage("BMT", BMT_SYNC_TIMEREQ_KEY..BMT_SEPARATOR..BMTBestUpdateTime, "GUILD");
	BMTTaskManager_Schedule(BMT_SYNC_TIME_DELAY, BMT_FinishSyncingTime);
end